/*************************************************************************
 *                                                                       * 
 *        N  A  S     P A R A L L E L     B E N C H M A R K S  3.3       *
 *                                                                       * 
 *                                  I S                                  * 
 *                                                                       * 
 ************************************************************************* 
 *                                                                       * 
 *   This benchmark is part of the NAS Parallel Benchmark 3.3 suite.     *
 *   It is described in NAS Technical Report 95-020.                     * 
 *                                                                       * 
 *   Permission to use, copy, distribute and modify this software        * 
 *   for any purpose with or without fee is hereby granted.  We          * 
 *   request, however, that all derived work reference the NAS           * 
 *   Parallel Benchmarks 3.3. This software is provided "as is"          *
 *   without express or implied warranty.                                * 
 *                                                                       * 
 *   Information on NPB 3.3, including the technical report, the         *
 *   original specifications, source code, results and information       * 
 *   on how to submit new results, is available at:                      * 
 *                                                                       * 
 *          http://www.nas.nasa.gov/Software/NPB                         * 
 *                                                                       * 
 *   Send comments or suggestions to  npb@nas.nasa.gov                   * 
 *   Send bug reports to              npb-bugs@nas.nasa.gov              * 
 *                                                                       * 
 *         NAS Parallel Benchmarks Group                                 * 
 *         NASA Ames Research Center                                     * 
 *         Mail Stop: T27A-1                                             * 
 *         Moffett Field, CA   94035-1000                                * 
 *                                                                       * 
 *         E-mail:  npb@nas.nasa.gov                                     * 
 *         Fax:     (650) 604-3957                                       * 
 *                                                                       * 
 ************************************************************************* 
 *                                                                       * 
 *   Author: M. Yarrow                                                   * 
 *           H. Jin                                                      * 
 *                                                                       * 
 *************************************************************************/

#include "is.h"
#include <cutil.h>
#include "cudaMPI.h"

#include "is_kernel.cu"

/********************/
/* MPI properties:  */
/********************/
int      my_rank,
         comm_size;


/********************/
/* Some global info */
/********************/
INT_TYPE *key_buff_ptr_global,         /* used by full_verify to get */
         total_local_keys,             /* copies of rank info        */
         total_lesser_keys;


int      passed_verification;
                                 


/************************************/
/* These are the three main arrays. */
/* See SIZE_OF_BUFFERS def above    */
/************************************/
INT_TYPE key_array[SIZE_OF_BUFFERS],    
         key_buff1[SIZE_OF_BUFFERS],    
         key_buff2[SIZE_OF_BUFFERS],
         bucket_size[NUM_BUCKETS+TEST_ARRAY_SIZE],     /* Top 5 elements for */
         bucket_size_totals[NUM_BUCKETS+TEST_ARRAY_SIZE], /* part. ver. vals */
         bucket_ptrs[NUM_BUCKETS],
         process_bucket_distrib_ptr1[NUM_BUCKETS+TEST_ARRAY_SIZE],   
         process_bucket_distrib_ptr2[NUM_BUCKETS+TEST_ARRAY_SIZE];   
int      send_count[MAX_PROCS], recv_count[MAX_PROCS],
         send_displ[MAX_PROCS], recv_displ[MAX_PROCS];


/**********************/
/* Partial verif info */
/**********************/
INT_TYPE2 test_index_array[TEST_ARRAY_SIZE],
         test_rank_array[TEST_ARRAY_SIZE],

         S_test_index_array[TEST_ARRAY_SIZE] = 
                             {48427,17148,23627,62548,4431},
         S_test_rank_array[TEST_ARRAY_SIZE] = 
                             {0,18,346,64917,65463},

         W_test_index_array[TEST_ARRAY_SIZE] = 
                             {357773,934767,875723,898999,404505},
         W_test_rank_array[TEST_ARRAY_SIZE] = 
                             {1249,11698,1039987,1043896,1048018},

         A_test_index_array[TEST_ARRAY_SIZE] = 
                             {2112377,662041,5336171,3642833,4250760},
         A_test_rank_array[TEST_ARRAY_SIZE] = 
                             {104,17523,123928,8288932,8388264},

         B_test_index_array[TEST_ARRAY_SIZE] = 
                             {41869,812306,5102857,18232239,26860214},
         B_test_rank_array[TEST_ARRAY_SIZE] = 
                             {33422937,10244,59149,33135281,99}, 

         C_test_index_array[TEST_ARRAY_SIZE] = 
                             {44172927,72999161,74326391,129606274,21736814},
         C_test_rank_array[TEST_ARRAY_SIZE] = 
                             {61147,882988,266290,133997595,133525895},

         D_test_index_array[TEST_ARRAY_SIZE] = 
                             {1317351170,995930646,1157283250,1503301535,1453734525},
         D_test_rank_array[TEST_ARRAY_SIZE] = 
                             {1,36538729,1978098519,2145192618,2147425337};

/********************/
/* CUDA Variables:  */
/********************/

INT_TYPE *key_array_CUDA;
INT_TYPE *bucket_size_CUDA;
INT_TYPE *bucket_size_totals_CUDA;
INT_TYPE *process_bucket_distrib_ptr1_CUDA;
INT_TYPE *process_bucket_distrib_ptr2_CUDA;
INT_TYPE2 *test_index_array_CUDA;
INT_TYPE *bucket_ptrs_CUDA;
INT_TYPE *key_buff1_CUDA;
INT_TYPE *send_displ_CUDA;
INT_TYPE *send_count_CUDA;
INT_TYPE *key_buff2_CUDA;
INT_TYPE2 *test_rank_array_CUDA;
int *passed_verification_CUDA;
int *key_buff_ptr_offset_CUDA;
INT_TYPE *total_local_keys_CUDA;
INT_TYPE *total_lesser_keys_CUDA;

float totalKernelLapse = 0.0, totalDMALapse = 0.0;
cudaEvent_t kernelStart, kernelStop, dmaStart, dmaStop;


/*
 *    FUNCTION RANDLC (X, A)
 *
 *  This routine returns a uniform pseudorandom double precision number in the
 *  range (0, 1) by using the linear congruential generator
 *
 *  x_{k+1} = a x_k  (mod 2^46)
 *
 *  where 0 < x_k < 2^46 and 0 < a < 2^46.  This scheme generates 2^44 numbers
 *  before repeating.  The argument A is the same as 'a' in the above formula,
 *  and X is the same as x_0.  A and X must be odd double precision integers
 *  in the range (1, 2^46).  The returned value RANDLC is normalized to be
 *  between 0 and 1, i.e. RANDLC = 2^(-46) * x_1.  X is updated to contain
 *  the new seed x_1, so that subsequent calls to RANDLC using the same
 *  arguments will generate a continuous sequence.
 *
 *  This routine should produce the same results on any computer with at least
 *  48 mantissa bits in double precision floating point data.  On Cray systems,
 *  double precision should be disabled.
 *
 *  David H. Bailey     October 26, 1990
 *
 *     IMPLICIT DOUBLE PRECISION (A-H, O-Z)
 *     SAVE KS, R23, R46, T23, T46
 *     DATA KS/0/
 *
 *  If this is the first call to RANDLC, compute R23 = 2 ^ -23, R46 = 2 ^ -46,
 *  T23 = 2 ^ 23, and T46 = 2 ^ 46.  These are computed in loops, rather than
 *  by merely using the ** operator, in order to insure that the results are
 *  exact on all systems.  This code assumes that 0.5D0 is represented exactly.
 */


/*****************************************************************/
/*************           R  A  N  D  L  C             ************/
/*************                                        ************/
/*************    portable random number generator    ************/
/*****************************************************************/

double	randlc( double *X, double *A )
{
      static int        KS=0;
      static double	R23, R46, T23, T46;
      double		T1, T2, T3, T4;
      double		A1;
      double		A2;
      double		X1;
      double		X2;
      double		Z;
      int     		i, j;

      if (KS == 0) 
      {
        R23 = 1.0;
        R46 = 1.0;
        T23 = 1.0;
        T46 = 1.0;
    
        for (i=1; i<=23; i++)
        {
          R23 = 0.50 * R23;
          T23 = 2.0 * T23;
        }
        for (i=1; i<=46; i++)
        {
          R46 = 0.50 * R46;
          T46 = 2.0 * T46;
        }
        KS = 1;
      }

/*  Break A into two parts such that A = 2^23 * A1 + A2 and set X = N.  */

      T1 = R23 * *A;
      j  = T1;
      A1 = j;
      A2 = *A - T23 * A1;

/*  Break X into two parts such that X = 2^23 * X1 + X2, compute
    Z = A1 * X2 + A2 * X1  (mod 2^23), and then
    X = 2^23 * Z + A2 * X2  (mod 2^46).                            */

      T1 = R23 * *X;
      j  = T1;
      X1 = j;
      X2 = *X - T23 * X1;
      T1 = A1 * X2 + A2 * X1;
      
      j  = R23 * T1;
      T2 = j;
      Z = T1 - T23 * T2;
      T3 = T23 * Z + A2 * X2;
      j  = R46 * T3;
      T4 = j;
      *X = T3 - T46 * T4;
      return(R46 * *X);
} 



/*****************************************************************/
/************   F  I  N  D  _  M  Y  _  S  E  E  D    ************/
/************                                         ************/
/************ returns parallel random number seq seed ************/
/*****************************************************************/

/*
 * Create a random number sequence of total length nn residing
 * on np number of processors.  Each processor will therefore have a 
 * subsequence of length nn/np.  This routine returns that random 
 * number which is the first random number for the subsequence belonging
 * to processor rank kn, and which is used as seed for proc kn ran # gen.
 */

double   find_my_seed( int  kn,       /* my processor rank, 0<=kn<=num procs */
                       int  np,       /* np = num procs                      */
                       long nn,       /* total num of ran numbers, all procs */
                       double s,      /* Ran num seed, for ex.: 314159265.00 */
                       double a )     /* Ran num gen mult, try 1220703125.00 */
{

  long   i;

  double t1,t2,t3,an;
  long   mq,nq,kk,ik;



      nq = nn / np;

      for( mq=0; nq>1; mq++,nq/=2 )
          ;

      t1 = a;

      for( i=1; i<=mq; i++ )
        t2 = randlc( &t1, &t1 );

      an = t1;

      kk = kn;
      t1 = s;
      t2 = an;

      for( i=1; i<=100; i++ )
      {
        ik = kk / 2;
        if( 2 * ik !=  kk ) 
            t3 = randlc( &t1, &t2 );
        if( ik == 0 ) 
            break;
        t3 = randlc( &t2, &t2 );
        kk = ik;
      }

      return( t1 );

}




/*****************************************************************/
/*************      C  R  E  A  T  E  _  S  E  Q      ************/
/*****************************************************************/

void	create_seq( double seed, double a )
{
	double x;
	int    i, k;

        k = MAX_KEY/4;

	for (i=0; i<NUM_KEYS; i++)
	{
	    x = randlc(&seed, &a);
	    x += randlc(&seed, &a);
    	    x += randlc(&seed, &a);
	    x += randlc(&seed, &a);  

            key_array[i] = k*x;
	}
}




/*****************************************************************/
/*************    F  U  L  L  _  V  E  R  I  F  Y     ************/
/*****************************************************************/


void full_verify( void )
{
    MPI_Status  status;
    MPI_Request request;
    
    INT_TYPE    i, j;
    INT_TYPE    k, last_local_key;

    
/*  Now, finally, sort the keys:  */
    for( i=0; i<total_local_keys; i++ )
        key_array[--key_buff_ptr_global[key_buff2[i]]-
                                 total_lesser_keys] = key_buff2[i];
    last_local_key = (total_local_keys<1)? 0 : (total_local_keys-1);

/*  Send largest key value to next processor  */
    if( my_rank > 0 )
        MPI_Irecv( &k,
                   1,
                   MP_KEY_TYPE,
                   my_rank-1,
                   1000,
                   MPI_COMM_WORLD,
                   &request );                   
    if( my_rank < comm_size-1 )
        MPI_Send( &key_array[last_local_key],
                  1,
                  MP_KEY_TYPE,
                  my_rank+1,
                  1000,
                  MPI_COMM_WORLD );
    if( my_rank > 0 )
        MPI_Wait( &request, &status );

/*  Confirm that neighbor's greatest key value 
    is not greater than my least key value       */              
    j = 0;
    if( my_rank > 0 && total_local_keys > 0 )
        if( k > key_array[0] )
            j++;


/*  Confirm keys correctly sorted: count incorrectly sorted keys, if any */
    for( i=1; i<total_local_keys; i++ )
        if( key_array[i-1] > key_array[i] )
            j++;


    if( j != 0 )
    {
        printf( "Processor %d:  Full_verify: number of keys out of sort: %d\n",
                my_rank, j );
    }
    else
        passed_verification++;
           

}

/*****************************************************************/
/*************       A L L T O A L L C O M M      ****************/
/*****************************************************************/

void alltoallcomm()
{

	INT_TYPE    i;

/*  This is the redistribution section:  first find out how many keys
    each processor will send to every other processor:                 */
    MPI_Alltoall( send_count,
                  1,
                  MPI_INT,
                  recv_count,
                  1,
                  MPI_INT,
                  MPI_COMM_WORLD );

/*  Determine the receive array displacements for the buckets */    
    recv_displ[0] = 0;
    for( i=1; i<comm_size; i++ )
        recv_displ[i] = recv_displ[i-1] + recv_count[i-1];


/*  Now send the keys to respective processors  */    
    MPI_Alltoallv( key_buff1,
                   send_count,
                   send_displ,
                   MP_KEY_TYPE,
                   key_buff2,
                   recv_count,
                   recv_displ,
                   MP_KEY_TYPE,
                   MPI_COMM_WORLD );

}

/*****************************************************************/
/*************             R  A  N  K  3  _  H  O  S  T                ****************/
/*****************************************************************/

void rank3_host( 
	int iteration,
	INT_TYPE *process_bucket_distrib_ptr1,
	INT_TYPE *process_bucket_distrib_ptr2,
	INT_TYPE *key_buff1,
	INT_TYPE *bucket_size_totals,
	INT_TYPE *key_buff2,
	INT_TYPE2 *test_rank_array,
	
	int *passed_verification,
	INT_TYPE **key_buff_ptr_global,
    INT_TYPE *total_local_keys,
    INT_TYPE *total_lesser_keys
	)
{

	INT_TYPE    i, k;
	INT_TYPE    shift = MAX_KEY_LOG_2 - NUM_BUCKETS_LOG_2;
	INT_TYPE    min_key_val, max_key_val;
	INT_TYPE2   j, m;
	INT_TYPE    *key_buff_ptr;
	
	int 	mpi_rank, mpi_node_count;

	/*  Get MPI rank and size */  
	MPI_Comm_rank( MPI_COMM_WORLD, &mpi_rank );
    MPI_Comm_size( MPI_COMM_WORLD, &mpi_node_count );

/*  The starting and ending bucket numbers on each processor are
    multiplied by the interval size of the buckets to obtain the 
    smallest possible min and greatest possible max value of any 
    key on each processor                                          */
    min_key_val = process_bucket_distrib_ptr1[mpi_rank] << shift;
    max_key_val = ((process_bucket_distrib_ptr2[mpi_rank] + 1) << shift)-1;
	
/*  Clear the work array */
    for( i=0; i<max_key_val-min_key_val+1; i++ )
        key_buff1[i] = 0;

/*  Determine the total number of keys on all other 
    processors holding keys of lesser value         */
    m = 0;
    for( k=0; k<mpi_rank; k++ )
        for( i= process_bucket_distrib_ptr1[k];
             i<=process_bucket_distrib_ptr2[k];
             i++ )  
            m += bucket_size_totals[i]; /*  m has total # of lesser keys */

/*  Determine total number of keys on this processor */
    j = 0;                                 
    for( i= process_bucket_distrib_ptr1[mpi_rank];
         i<=process_bucket_distrib_ptr2[mpi_rank];
         i++ )  
        j += bucket_size_totals[i];     /* j has total # of local keys   */


/*  Ranking of all keys occurs in this section:                 */
/*  shift it backwards so no subtractions are necessary in loop */
    key_buff_ptr = key_buff1 - min_key_val;

/*  In this section, the keys themselves are used as their 
    own indexes to determine how many of each there are: their
    individual population                                       */
    for( i=0; i<j; i++ )
        key_buff_ptr[key_buff2[i]]++;  /* Now they have individual key   */
                                       /* population                     */

/*  To obtain ranks of each key, successively add the individual key
    population, not forgetting the total of lesser keys, m.
    NOTE: Since the total of lesser keys would be subtracted later 
    in verification, it is no longer added to the first key population 
    here, but still needed during the partial verify test.  This is to 
    ensure that 32-bit key_buff can still be used for class D.           */
/*    key_buff_ptr[min_key_val] += m;    */
    for( i=min_key_val; i<max_key_val; i++ )   
        key_buff_ptr[i+1] += key_buff_ptr[i];  


/* This is the partial verify test section */
/* Observe that test_rank_array vals are   */
/* shifted differently for different cases */
    for( i=0; i<TEST_ARRAY_SIZE; i++ )
    {                                             
        k = bucket_size_totals[i+NUM_BUCKETS];    /* Keys were hidden here */
        if( min_key_val <= k  &&  k <= max_key_val )
        {
            /* Add the total of lesser keys, m, here */
            INT_TYPE2 key_rank = key_buff_ptr[k-1] + m;
            int failed = 0;

            switch( CLASS )
            {
                case 'S':
                    if( i <= 2 )
                    {
                        if( key_rank != test_rank_array[i]+iteration )
                            failed = 1;
                        else
                            (*passed_verification)++;
                    }
                    else
                    {
                        if( key_rank != test_rank_array[i]-iteration )
                            failed = 1;
                        else
                            (*passed_verification)++;
                    }
                    break;
                case 'W':
                    if( i < 2 )
                    {
                        if( key_rank != test_rank_array[i]+(iteration-2) )
                            failed = 1;
                        else
                            (*passed_verification)++;
                    }
                    else
                    {
                        if( key_rank != test_rank_array[i]-iteration )
                            failed = 1;
                        else
                            (*passed_verification)++;
                    }
                    break;
                case 'A':
                    if( i <= 2 )
        	    {
                        if( key_rank != test_rank_array[i]+(iteration-1) )
                            failed = 1;
                        else
                            (*passed_verification)++;
        	    }
                    else
                    {
                        if( key_rank != test_rank_array[i]-(iteration-1) )
                            failed = 1;
                        else
                            (*passed_verification)++;
                    }
                    break;
                case 'B':
                    if( i == 1 || i == 2 || i == 4 )
        	    {
                        if( key_rank != test_rank_array[i]+iteration )
                            failed = 1;
                        else
                            (*passed_verification)++;
        	    }
                    else
                    {
                        if( key_rank != test_rank_array[i]-iteration )
                            failed = 1;
                        else
                            (*passed_verification)++;
                    }
                    break;
                case 'C':
                    if( i <= 2 )
        	    {
                        if( key_rank != test_rank_array[i]+iteration )
                            failed = 1;
                        else
                            (*passed_verification)++;
        	    }
                    else
                    {
                        if( key_rank != test_rank_array[i]-iteration )
                            failed = 1;
                        else
                            (*passed_verification)++;
                    }
                    break;
                case 'D':
                    if( i < 2 )
        	    {
                        if( key_rank != test_rank_array[i]+iteration )
                            failed = 1;
                        else
                            (*passed_verification)++;
        	    }
                    else
                    {
                        if( key_rank != test_rank_array[i]-iteration )
                            failed = 1;
                        else
                            (*passed_verification)++;
                    }
                    break;
            }
            if( failed == 1 )
                printf( "Failed partial verification: "
                        "iteration %d, processor %d, test key %d\n", 
                         iteration, mpi_rank, (int)i );
        }
    }




/*  Make copies of rank info for use by full_verify: these variables
    in rank are local; making them global slows down the code, probably
    since they cannot be made register by compiler                        */

    if( iteration == MAX_ITERATIONS ) 
    {
        *key_buff_ptr_global = key_buff_ptr;		// Set the contents of key_buff_ptr_global equal to the location of key_buff_ptr
		//printf("key_buff_ptr_offset = %ld.\n", key_buff1 - key_buff_ptr);
        *total_local_keys    = j;
        *total_lesser_keys   = 0;  /* no longer set to 'm', see note above */
    }

}     


/*****************************************************************/
/*************           C U D A   A L L O C A T I O N           ****************/
/*****************************************************************/

void cuda_allocation(){

	// Allocate key_array
	CUDA_SAFE_CALL( cudaMalloc((void**) &key_array_CUDA, SIZE_OF_BUFFERS*sizeof(INT_TYPE)) );
	
	// Allocate bucket_size	
	CUDA_SAFE_CALL( cudaMalloc((void**) &bucket_size_CUDA, (NUM_BUCKETS+TEST_ARRAY_SIZE)*sizeof(INT_TYPE)) );
	
	// Allocate bucket_size_totals	
	CUDA_SAFE_CALL( cudaMalloc((void**) &bucket_size_totals_CUDA, (NUM_BUCKETS+TEST_ARRAY_SIZE)*sizeof(INT_TYPE)) );
	
	// Allocate process_bucket_distrib_ptr1	
	CUDA_SAFE_CALL( cudaMalloc((void**) &process_bucket_distrib_ptr1_CUDA, (NUM_BUCKETS+TEST_ARRAY_SIZE)*sizeof(INT_TYPE)) );
	
	// Allocate process_bucket_distrib_ptr2	
	CUDA_SAFE_CALL( cudaMalloc((void**) &process_bucket_distrib_ptr2_CUDA, (NUM_BUCKETS+TEST_ARRAY_SIZE)*sizeof(INT_TYPE)) );
	
	// Allocate test_index_array	
	CUDA_SAFE_CALL( cudaMalloc((void**) &test_index_array_CUDA, TEST_ARRAY_SIZE*sizeof(INT_TYPE2)) );
	
	// Allocate bucket_ptrs	
	CUDA_SAFE_CALL( cudaMalloc((void**) &bucket_ptrs_CUDA, NUM_BUCKETS*sizeof(INT_TYPE)) );
	
	// Allocate key_buff1	
	CUDA_SAFE_CALL( cudaMalloc((void**) &key_buff1_CUDA, SIZE_OF_BUFFERS*sizeof(INT_TYPE)) );
	
	// Allocate send_displ	
	CUDA_SAFE_CALL( cudaMalloc((void**) &send_displ_CUDA, MAX_PROCS*sizeof(INT_TYPE)) );
	
	// Allocate send_count	
	CUDA_SAFE_CALL( cudaMalloc((void**) &send_count_CUDA, MAX_PROCS*sizeof(INT_TYPE)) );

}

/*****************************************************************/
/*************           C U D A   F R E E                                ****************/
/*****************************************************************/

void cuda_free(){

	// Free key_array
	CUDA_SAFE_CALL(cudaFree(key_array_CUDA));
	
	// Free bucket_size
	CUDA_SAFE_CALL(cudaFree(bucket_size_CUDA));
	
	// Free bucket_size_totals
	CUDA_SAFE_CALL(cudaFree(bucket_size_totals_CUDA));
	
	// Free process_bucket_distrib_ptr1
	CUDA_SAFE_CALL(cudaFree(process_bucket_distrib_ptr1_CUDA));
	
	// Free process_bucket_distrib_ptr2
	CUDA_SAFE_CALL(cudaFree(process_bucket_distrib_ptr2_CUDA));
	
	// Free test_index_array
	CUDA_SAFE_CALL(cudaFree(test_index_array_CUDA));
	
	// Free bucket_ptrs
	CUDA_SAFE_CALL(cudaFree(bucket_ptrs_CUDA));
	
	// Free key_buff1
	CUDA_SAFE_CALL(cudaFree(key_buff1_CUDA));
	
	// Free send_displ
	CUDA_SAFE_CALL(cudaFree(send_displ_CUDA));
	
	// Free send_count
	CUDA_SAFE_CALL(cudaFree(send_count_CUDA));

}

/*****************************************************************/
/*************             R  A  N  K             ****************/
/*****************************************************************/

void rank( int iteration, int warmup )
{
	float thisLapse;
	
	// ============================
	// Perform rank1 initial in-DMA
	// ============================
	
	// Record dma start time
	cudaEventRecord(dmaStart, 0);
	
	// in-DMA key_array
	CUDA_SAFE_CALL( cudaMemcpy(key_array_CUDA, key_array, SIZE_OF_BUFFERS*sizeof(INT_TYPE), cudaMemcpyHostToDevice) );
	
	// in-DMA test_index_array
	CUDA_SAFE_CALL( cudaMemcpy(test_index_array_CUDA, test_index_array, TEST_ARRAY_SIZE*sizeof(INT_TYPE2), cudaMemcpyHostToDevice) );
	
	// in-DMA bucket_ptrs
	CUDA_SAFE_CALL( cudaMemcpy(bucket_ptrs_CUDA, bucket_ptrs, NUM_BUCKETS*sizeof(INT_TYPE), cudaMemcpyHostToDevice) );
	
	// in-DMA key_buff1
	CUDA_SAFE_CALL( cudaMemcpy(key_buff1_CUDA, key_buff1, SIZE_OF_BUFFERS*sizeof(INT_TYPE), cudaMemcpyHostToDevice) );
	
	// in-DMA send_displ - only need for rank2 kernel
	CUDA_SAFE_CALL( cudaMemcpy(send_displ_CUDA, send_displ, MAX_PROCS*sizeof(INT_TYPE), cudaMemcpyHostToDevice) );
	
	// in-DMA send_count - only need for rank2 kernel
	CUDA_SAFE_CALL( cudaMemcpy(send_count_CUDA, send_count, MAX_PROCS*sizeof(INT_TYPE), cudaMemcpyHostToDevice) );
	
	// Record dma stop time and synchronize
	cudaEventRecord(dmaStop, 0);
	cudaEventSynchronize(dmaStop);
	
	// Add to DMA lapse count
	if(warmup != 1){
		thisLapse = 0.0;
		cudaEventElapsedTime(&thisLapse, dmaStart, dmaStop);	
		//printf("Rank(%d) DMA-in lapse = %12.2f ms.\n", iteration, thisLapse);
		totalDMALapse += thisLapse;
	}
	
	// ============================
	// Call rank1 kernels
	// ============================
	
	// Record kernel start time
	cudaEventRecord(kernelStart, 0);
	
	// Call initialization kernel
	rank1_kernel_init<<<2,512>>>( 
		my_rank,
		iteration,
		key_array_CUDA,
		bucket_size_CUDA,
		bucket_size_totals_CUDA,
		process_bucket_distrib_ptr1_CUDA,
		process_bucket_distrib_ptr2_CUDA,
		test_index_array_CUDA
	);
	
	// Block until the device has completed
	cudaThreadSynchronize();
	
	// Call key counting kernel
	rank1_kernel_countkeys<<<2048,512>>>( 
		key_array_CUDA,
		bucket_size_CUDA
	);
	
	// Block until the device has completed
	cudaThreadSynchronize();
	
	// Call bucket size kernel
	rank1_kernel_bucketsizes<<<1,512>>>( 
		bucket_size_CUDA,
		bucket_ptrs_CUDA
	);
	
	// Block until the device has completed
	cudaThreadSynchronize();

	// Call sorting kernel
	rank1_kernel_sort<<<2048,512>>>( 
		key_array_CUDA,
		bucket_size_CUDA,
		bucket_ptrs_CUDA,
		key_buff1_CUDA
	);
	
	// Block until the device has completed
	cudaThreadSynchronize();
	
	// Record kernel stop time and synchronize
	cudaEventRecord(kernelStop, 0);
	cudaEventSynchronize(kernelStop);
	
	// Add to kernel lapse count
	thisLapse = 0.0;
	cudaEventElapsedTime(&thisLapse, kernelStart, kernelStop);
	if(warmup != 1){
		//printf("Rank1(%d) kernel lapse = %12.2f ms.\n", iteration, thisLapse);
		totalKernelLapse += thisLapse;
	}
	
	// ============================
	// Handle timers
	// ============================
	
#ifdef  TIMING_ENABLED
    timer_stop( 2 );
    timer_start( 3 );
#endif

	// ============================
	// Perform MPI with on-device memory
	// ============================
	
    cudaMPI_Allreduce( bucket_size_CUDA, 
                   bucket_size_totals_CUDA, 
                   NUM_BUCKETS+TEST_ARRAY_SIZE, 
                   MP_KEY_TYPE,
                   MPI_SUM,
                   MPI_COMM_WORLD );
				   
	// ============================
	// Handle timers
	// ============================
	
#ifdef  TIMING_ENABLED
    timer_stop( 3 );
    timer_start( 2 );
#endif
				   
	// ============================
	// Call rank2 kernel
	// ============================
	
	// Record start time
	cudaEventRecord(kernelStart, 0);
	
	rank2_kernel<<<1,128>>>( 
		comm_size,
		bucket_size_CUDA,
		bucket_size_totals_CUDA,
		process_bucket_distrib_ptr1_CUDA,
		process_bucket_distrib_ptr2_CUDA,
		send_displ_CUDA,
		send_count_CUDA	
	);	
	
	// Record kernel stop time and synchronize
	cudaEventRecord(kernelStop, 0);
	cudaEventSynchronize(kernelStop);
	
	// Add to kernel lapse count
	if(warmup != 1){
		thisLapse = 0.0;
		cudaEventElapsedTime(&thisLapse, kernelStart, kernelStop);	
		//printf("Rank2(%d) kernel lapse = %12.2f ms.\n", iteration, thisLapse);
		totalKernelLapse += thisLapse;
	}
	
	// ============================
	// Perform rank1/2 out-DMA
	// ============================
	
	// Record dma start time
	cudaEventRecord(dmaStart, 0);
	
	// out-DMA key_array
	CUDA_SAFE_CALL( cudaMemcpy(key_array, key_array_CUDA, SIZE_OF_BUFFERS*sizeof(INT_TYPE), cudaMemcpyDeviceToHost) );
	
	// out-DMA bucket_size
	CUDA_SAFE_CALL( cudaMemcpy(bucket_size, bucket_size_CUDA, (NUM_BUCKETS+TEST_ARRAY_SIZE)*sizeof(INT_TYPE), cudaMemcpyDeviceToHost) );
	
	// out-DMA bucket_size_totals
	CUDA_SAFE_CALL( cudaMemcpy(bucket_size_totals, bucket_size_totals_CUDA, (NUM_BUCKETS+TEST_ARRAY_SIZE)*sizeof(INT_TYPE), cudaMemcpyDeviceToHost) );
	
	// out-DMA process_bucket_distrib_ptr1 - only need for rank2 kernel
	CUDA_SAFE_CALL( cudaMemcpy(process_bucket_distrib_ptr1, process_bucket_distrib_ptr1_CUDA, (NUM_BUCKETS+TEST_ARRAY_SIZE)*sizeof(INT_TYPE), cudaMemcpyDeviceToHost) );
	
	// out-DMA process_bucket_distrib_ptr2 - only need for rank2 kernel
	CUDA_SAFE_CALL( cudaMemcpy(process_bucket_distrib_ptr2, process_bucket_distrib_ptr2_CUDA, (NUM_BUCKETS+TEST_ARRAY_SIZE)*sizeof(INT_TYPE), cudaMemcpyDeviceToHost) );
	
	// out-DMA test_index_array
	CUDA_SAFE_CALL( cudaMemcpy(test_index_array, test_index_array_CUDA, TEST_ARRAY_SIZE*sizeof(INT_TYPE2), cudaMemcpyDeviceToHost) );
	
	// out-DMA key_buff1
	CUDA_SAFE_CALL( cudaMemcpy(key_buff1, key_buff1_CUDA, SIZE_OF_BUFFERS*sizeof(INT_TYPE), cudaMemcpyDeviceToHost) );
	
	// out-DMA send_displ - only need for rank2 kernel
	CUDA_SAFE_CALL( cudaMemcpy(send_displ, send_displ_CUDA, MAX_PROCS*sizeof(INT_TYPE), cudaMemcpyDeviceToHost) );
	
	// out-DMA send_count - only need for rank2 kernel
	CUDA_SAFE_CALL( cudaMemcpy(send_count, send_count_CUDA, MAX_PROCS*sizeof(INT_TYPE), cudaMemcpyDeviceToHost) );
	
	// Record DMA stop time and synchronize
	cudaEventRecord(dmaStop, 0);
	cudaEventSynchronize(dmaStop);
	
	// Add to DMA lapse count
	if(warmup != 1){
		thisLapse = 0.0;
		cudaEventElapsedTime(&thisLapse, dmaStart, dmaStop);
		//printf("Rank(%d) DMA-out lapse = %12.2f ms.\n", iteration, thisLapse);
		totalDMALapse += thisLapse;
	}
	
	// ============================
	// Handle timers
	// ============================
	
#ifdef  TIMING_ENABLED
    timer_stop( 2 );
    timer_start( 3 ); 
#endif
	
	// ============================
	// Perform alltoall communication
	// ============================
	
	alltoallcomm();

	// ============================
	// Handle timers
	// ============================
	
#ifdef  TIMING_ENABLED
    timer_stop( 3 ); 
    timer_start( 2 );
#endif
	
	// ============================
	// Call rank3 on host
	// ============================
	
	rank3_host(
		iteration,
		process_bucket_distrib_ptr1,
		process_bucket_distrib_ptr2,
		key_buff1,
		bucket_size_totals,
		key_buff2,
		test_rank_array,		
		&passed_verification,
		&key_buff_ptr_global,
		&total_local_keys,
		&total_lesser_keys
	);
	
}      

/*****************************************************************/
/*************             M  A  I  N             ****************/
/*****************************************************************/

int main( int argc, char **argv )
{

    int             i, iteration, itemp;

    double          timecounter, maxtime;
	double			maxKernelLapse, maxDMALapse;


/*  Initialize MPI */
    MPI_Init( &argc, &argv );
    MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );
    MPI_Comm_size( MPI_COMM_WORLD, &comm_size );

/*  Initialize CUDA */
	CUT_DEVICE_INIT(argc, argv);

/*  Initialize the verification arrays if a valid class */
    for( i=0; i<TEST_ARRAY_SIZE; i++ )
        switch( CLASS )
        {
            case 'S':
                test_index_array[i] = S_test_index_array[i];
                test_rank_array[i]  = S_test_rank_array[i];
                break;
            case 'A':
                test_index_array[i] = A_test_index_array[i];
                test_rank_array[i]  = A_test_rank_array[i];
                break;
            case 'W':
                test_index_array[i] = W_test_index_array[i];
                test_rank_array[i]  = W_test_rank_array[i];
                break;
            case 'B':
                test_index_array[i] = B_test_index_array[i];
                test_rank_array[i]  = B_test_rank_array[i];
                break;
            case 'C':
                test_index_array[i] = C_test_index_array[i];
                test_rank_array[i]  = C_test_rank_array[i];
                break;
            case 'D':
                test_index_array[i] = D_test_index_array[i];
                test_rank_array[i]  = D_test_rank_array[i];
                break;
        };

        

/*  Printout initial NPB info */
    if( my_rank == 0 )
    {
        printf( "\n\n NAS Parallel Benchmarks 3.3 -- IS Benchmark\n\n" );
        printf( " Size:  %ld  (class %c)\n", (long)TOTAL_KEYS*MIN_PROCS, CLASS );
        printf( " Iterations:   %d\n", MAX_ITERATIONS );
        printf( " Number of processes:     %d\n", comm_size );
    }

/*  Check that actual and compiled number of processors agree */
    if( comm_size != NUM_PROCS )
    {
        if( my_rank == 0 )
            printf( "\n ERROR: compiled for %d processes\n"
                    " Number of active processes: %d\n"
                    " Exiting program!\n\n", NUM_PROCS, comm_size );
        MPI_Finalize();
        exit( 1 );
    }

/*  Check to see whether total number of processes is within bounds.
    This could in principle be checked in setparams.c, but it is more
    convenient to do it here                                               */
    if( comm_size < MIN_PROCS || comm_size > MAX_PROCS)
    {
       if( my_rank == 0 )
           printf( "\n ERROR: number of processes %d not within range %d-%d"
                   "\n Exiting program!\n\n", comm_size, MIN_PROCS, MAX_PROCS);
       MPI_Finalize();
       exit( 1 );
    }


/*  Generate random number sequence and subsequent keys on all procs */
    create_seq( find_my_seed( my_rank, 
                              comm_size, 
                              4*(long)TOTAL_KEYS*MIN_PROCS,
                              314159265.00,      /* Random number gen seed */
                              1220703125.00 ),   /* Random number gen mult */
                1220703125.00 );                 /* Random number gen mult */

/* Allocate CUDA memory */
	cuda_allocation();

/*  Do one interation for free (i.e., untimed) to guarantee initialization of  
    all data and code pages and respective tables */
    rank( 1, 1 );  

/*  Start verification counter */
    passed_verification = 0;

    if( my_rank == 0 && CLASS != 'S' ) printf( "\n   iteration\n" );

/*  Initialize overall timer  */             
    timer_clear( 0 );

/*  Initialize separate communication, computation timing */
#ifdef  TIMING_ENABLED 
    for( i=1; i<=3; i++ ) timer_clear( i );
#endif

/* Initialize CUDA timers */
	cudaEventCreate(&kernelStart);
	cudaEventCreate(&kernelStop);
	cudaEventCreate(&dmaStart);
	cudaEventCreate(&dmaStop);

/*  Start overall timer  */             
    timer_start( 0 );
	
#ifdef  TIMING_ENABLED
    timer_start( 1 );
    timer_start( 2 );
#endif


/*  This is the main iteration */
    for( iteration=1; iteration<=MAX_ITERATIONS; iteration++ )
    {
		// Print
        if( my_rank == 0 && CLASS != 'S' ) 
			printf( "        %d\n", iteration );
		
		// Call rank
		rank(iteration, 0);
		
    }

#ifdef  TIMING_ENABLED
    timer_stop( 2 );
    timer_stop( 1 );
#endif

/*  Stop timer, obtain time for processors */
    timer_stop( 0 );

    timecounter = timer_read( 0 );

/*  End of timing, obtain maximum time of all processors */
    MPI_Reduce( &timecounter,
                &maxtime,
                1,
                MPI_DOUBLE,
                MPI_MAX,
                0,
                MPI_COMM_WORLD );

	double totalKernelLapse_dbl = (double) totalKernelLapse;
	MPI_Reduce( &totalKernelLapse_dbl,
                &maxKernelLapse,
                1,
                MPI_DOUBLE,
                MPI_MAX,
                0,
                MPI_COMM_WORLD );
	
	double totalDMALapse_dbl = (double) totalDMALapse;
	MPI_Reduce( &totalDMALapse_dbl,
                &maxDMALapse,
                1,
                MPI_DOUBLE,
                MPI_MAX,
                0,
                MPI_COMM_WORLD );
				
#ifdef  TIMING_ENABLED
    {
        double    tmin, tsum, tmax;
    
        if( my_rank == 0 )
        {
            printf( "\ntimer 1/2/3 = total/computation/communication time\n");
            printf( "              min                avg                max\n" );
        }
        for( i=1; i<=3; i++ )
        {
            timecounter = timer_read( i );
            MPI_Reduce( &timecounter,
                        &tmin,
                        1,
                        MPI_DOUBLE,
                        MPI_MIN,
                        0,
                        MPI_COMM_WORLD );
            MPI_Reduce( &timecounter,
                        &tsum,
                        1,
                        MPI_DOUBLE,
                        MPI_SUM,
                        0,
                        MPI_COMM_WORLD );
            MPI_Reduce( &timecounter,
                        &tmax,
                        1,
                        MPI_DOUBLE,
                        MPI_MAX,
                        0,
                        MPI_COMM_WORLD );
            if( my_rank == 0 )
                printf( "timer %d:    %f           %f            %f\n",
                        i, tmin, tsum/((double) comm_size), tmax );
        }
        if( my_rank == 0 )
            printf( "\n" );
    }
#endif

/* Barrier so no one leaves */
	MPI_Barrier(MPI_COMM_WORLD);
	
/*  This tests that keys are in sequence: sorting of last ranked key seq
    occurs here, but is an untimed operation                             */	
    full_verify();

/* Barrier so no one leaves */
	MPI_Barrier(MPI_COMM_WORLD);
	
/*  Obtain verification counter sum */
    itemp = passed_verification;
    MPI_Reduce( &itemp,
                &passed_verification,
                1,
                MPI_INT,
                MPI_SUM,
                0,
                MPI_COMM_WORLD );

/*  The final printout  */
    if( my_rank == 0 )
    {
        if( passed_verification != 5*MAX_ITERATIONS + comm_size )
            passed_verification = 0;
        
		// Print results
		c_print_results( "IS",
                         CLASS,
                         (int)(TOTAL_KEYS),
                         MIN_PROCS,
                         0,
                         MAX_ITERATIONS,
                         NUM_PROCS,
                         comm_size,
                         maxtime,
                         ((double) (MAX_ITERATIONS)*TOTAL_KEYS*MIN_PROCS)
                                                      /maxtime/1000000.,
                         "keys ranked", 
                         passed_verification,
                         NPBVERSION,
                         COMPILETIME,
                         MPICC,
                         CLINK,
                         CMPI_LIB,
                         CMPI_INC,
                         CFLAGS,
                         CLINKFLAGS );
		
		// Print kernel time
		printf("Total kernel time: %12.2f s\n", maxKernelLapse/1000);
		printf("Total DMA time: %12.2f s\n", maxDMALapse/1000);
		
    }
	
/* Free CUDA memory */
	cuda_free();

/* Free CUDA events */
	cudaEventDestroy(kernelStart);
	cudaEventDestroy(kernelStop);
	cudaEventDestroy(dmaStart);
	cudaEventDestroy(dmaStop);
	
/* CUDA Finalize */
	CUT_EXIT(argc, argv);
	
/* MPI Finalize */	
    MPI_Finalize();

    return 0;
	
         /**************************/
}        /*  E N D  P R O G R A M  */
         /**************************/

		 
		 
